#!/usr/bin/env python

## @file
# Clean the cluster nodes after REPET was used.
#
# usage: CleanClusterNodesAfterRepet.py [ options ]
# options:
#      -h: this help
#      -n: node to clean (otherwise all nodes will be cleaned)
#      -t: temporary directory (e.g. '/tmp')
#      -p: pattern (e.g. 'DmelChr4*')
#      -v: verbosity level (default=0/1/2)

import os
import sys
import getopt
import exceptions
import threading
import glob
import stat
import time
import pwd

class CleanClusterNodesAfterRepet( object ):
    """
    Clean the cluster nodes after REPET was used.
    """
    
    def __init__( self ):
        """
        Constructor.
        """
        self._node = ""
        self._tmpDir = ""
        self._pattern = ""
        self._verbose = 0

    def help( self ):
        """
        Display the help on stdout.
        """
        print
        print "usage: %s.py [ options ]" % ( type(self).__name__ )
        print "options:"
        print "     -h: this help"
        print "     -n: node to clean (otherwise all nodes will be cleaned)"
        print "     -t: temporary directory (e.g. '/tmp')"
        print "     -p: pattern (e.g. 'DmelChr4*')"
        print "     -v: verbosity level (default=0/1/2)"
        print
        
    def setAttributesFromCmdLine( self ):
        """
        Set the attributes from the command-line.
        """
        try:
            opts, args = getopt.getopt(sys.argv[1:],"hi:n:t:p:v:")
        except getopt.GetoptError, err:
            print str(err); self.help(); sys.exit(1)
        for o,a in opts:
            if o == "-h":
                self.help(); sys.exit(0)
            elif o == "-n":
                self.setNode( a )
            elif o == "-t":
                self.setTempDirectory( a )
            elif o == "-p":
                self.setPattern( a )
            elif o == "-v":
                self.setVerbosityLevel( a )
                
    def setNode( self, a ):
        self._node = a
        
    def setTempDirectory( self, a ):
        if a[-1] == "/":
            self._tmpDir = a[:-1]
        else:
            self._tmpDir = a
            
    def setPattern( self, a ):
        self._pattern = a
        
    def setVerbosityLevel( self, verbose ):
        self._verbose = int(verbose)
        
    def checkAttributes( self ):
        """
        Before running, check the required attributes are properly filled.
        """
        if self._tmpDir == "":
            print "ERROR: need a valid temporary directory"
            self.help(); sys.exit(1)
            
    def getNodeList( self ):
        """
        Return the list of the names of each node.
        """
        lNodes = []
        log = os.system( "qhost > qhost.txt" )
        if log != 0: print "ERROR with qhost"; sys.exit(1)
        inF = open( "qhost.txt", "r" )
        line = inF.readline()
        line = inF.readline()
        line = inF.readline()
        while True:
            if line == "":
                break
            tokens = line.split()
            if tokens[3] == "-":
                line = inF.readline()
                continue
            lNodes.append( tokens[0] )
            line = inF.readline()
        return lNodes
    
    def showNodeList( self, lNodes ):
        print "nb of nodes: %i" % ( len(lNodes) )
        for i in range(0,len(lNodes)):
            print " %i: %s" % ( i+1, lNodes[i] )
            
    def cleanNodes( self, lNodes ):
        """
        Connect to each job and clean the temporary directory.
        """
        nbNodes = len(lNodes)
        nodeCount = 0
        for node in lNodes:
            nodeCount += 1
            if self._verbose > 0:
                print "connect to node '%s' (%i/%i)..." % ( node, nodeCount, nbNodes )
                sys.stdout.flush()
            cmd = "ssh"
            cmd += " -q %s " % ( node )
            cmd += "'find %s" % ( self._tmpDir )
            cmd += " -user %s" % ( os.environ["USER"] )
            if self._pattern != "":
                cmd += " -name '%s'" % ( self._pattern )
            cmd += " 2> /dev/null -exec rm -rf {} \; ; exit'"
            if self._verbose > 0: print cmd; sys.stdout.flush()
            log = os.system( cmd )  # warning, even if everything goes right, ssh returns an error code, i.e. different than 0
##             tcn = ThreadCleanNode( self._tmpDir, node )
##             tcn.start()
##             tcn.join()
    
    def clean( self ):
        if os.path.exists( "qhost.txt" ):
            os.remove( "qhost.txt" )
            
    def start( self ):
        """
        Useful commands before running the program.
        """
        if self._verbose > 0:
            print "START %s" % ( type(self).__name__ ); sys.stdout.flush()
        self.checkAttributes()        
        
    def end( self ):
        """
        Useful commands before ending the program.
        """
        self.clean()
        if self._verbose > 0:
            print "END %s" % ( type(self).__name__ ); sys.stdout.flush()
            
    def run( self ):
        """
        Run the program.
        """
        self.start()
        if self._node == "":
            lNodes = self.getNodeList()
        else:
            lNodes = [ self._node ]
        if self._verbose > 0: self.showNodeList( lNodes )
        self.cleanNodes( lNodes )
        self.end()
        
        
class ThreadCleanNode( threading.Thread ):
    """
    Clean a specific node via a thread.
    """
    def __init__( self, tmpDir, node ):
        threading.Thread.__init__( self )
        self._tmpDir = tmpDir
        self._node = node
    def run( self ):
        pass
##        os.system( "qrsh -l hostname=%s" % ( self._node ) )
##         if os.path.exists( self._tmpDir ):
##             lTmpDir = glob.glob( "%s/*" % ( self._tmpDir ) )
##             for i in lTmpDir:
##                 s = os.stat( i )
##                 if not stat.S_ISDIR( s[stat.ST_MODE] ) \
##                        and self.ownByMe( s[stat.ST_UID] ) \
##                        and self.isOldEnough( s[stat.ST_MTIME] ):
##                     os.remove( i )
##         os.system( "exit" )
    def ownByMe( self, fileUid ):
        if pwd.getpwuid( fileUid )[0] == os.environ["USER"]:
            return True
        return False
    def isOldEnough( self, fileMtime ):
        return False


if __name__ == "__main__":
    i = CleanClusterNodesAfterRepet()
    i.setAttributesFromCmdLine()
    i.run()
