#!/usr/bin/env python

##@file
# Get the sequences corresponding to input coordinates


import os
import sys
import getopt
import time
from pyRepetUnit.commons.sql.DbMySql import DbMySql
from pyRepetUnit.commons.sql.TablePathAdaptator import TablePathAdaptator
from pyRepetUnit.commons.sql.TableSeqAdaptator import TableSeqAdaptator
from pyRepetUnit.commons.coord.PathUtils import PathUtils
from pyRepetUnit.commons.seq.Bioseq import Bioseq
from pyRepetUnit.commons.coord.SetUtils import SetUtils


## Get the sequences corresponding to input coordinates
#
class GetSequencesFromAnnotations( object ):
    
    def __init__( self ):
        """
        Constructor.
        """
        self._inCoord = ""
        self._formatInCoord = "path"
        self._inGenome = ""
        self._inRefseqs = ""
        self._refseqName = ""
        self._data = "matches"
        self._minProp = 0.0
        self._getAnnot = False
        self._configFile = ""
        self._verbose = 0
        self._typeInCoord = "table"
        self._typeInGenome = "table"
        self._typeInRefseqs = "table"
        self._iDb = None
        self._tpa = TablePathAdaptator()
        self._genomeAdaptator = TableSeqAdaptator()
        self._refseqsAdaptator = TableSeqAdaptator()
        self._outPathFile = ""
        self._outPathHandler = None
        
        
    def help( self ):
        """
        Display the help on stdout.
        """
        print
        print "usage: GetSequencesFromAnnotations.py [ options ]"
        print "options:"
        print "     -h: this help"
        print "     -i: input data with coordinates (file or table)"
        print "     -f: format of the input coordinates (default='path')"
        print "     -g: input data with genomic sequences (file or table, format='fasta', queries in 'path')"
        print "     -s: input data with the reference sequences (file or table, format='fasta')"
        print "     -r: name of a reference sequence (subject in 'path', all by default)"
        print "     -d: sequences to retrieve (default='matches'/'chains')"
        print "     -y: minimum copy proportion compare to reference sequences (default=0.0)"
        print "     -a: also retrieve the annotations (format='path')"
        print "     -C: configuration file"
        print "     -v: verbosity level (default=0/1/2)"
        print
        
        
    def setAttributesFromCmdLine( self ):
        """
        Set the attributes from the command-line.
        """
        try:
            opts, args = getopt.getopt(sys.argv[1:],"hi:f:g:s:r:d:y:aC:v:")
        except getopt.GetoptError, err:
            print str(err); self.help(); sys.exit(1)
        for o,a in opts:
            if o == "-h":
                self.help(); sys.exit(0)
            elif o == "-i":
                self._inCoord = a
            elif o == "-f":
                self._formatInCoord = a
            elif o == "-g":
                self._inGenome = a
            elif o == "-s":
                self._inRefseqs = a
            elif o == "-r":
                self._refseqName = a
            elif o == "-d":
                self._data = a
            elif o == "-y":
                self._minProp = float(a)
            elif o == "-a":
                self._getAnnot = True
            elif o == "-C":
                self._configFile = a
            elif o == "-v":
                self._verbose = int(a)
                
                
    def checkAttributes( self ):
        """
        Check the attributes are valid before running the algorithm.
        """
        if self._configFile == "":
            msg = "ERROR: missing configuration file (-C)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
        if not os.path.exists( self._configFile ):
            msg = "ERROR: configuration file '%s' doesn't exist" % ( self._configFile )
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
        self._iDb = DbMySql( cfgFileName=self._configFile )
        if self._inCoord == "":
            msg = "ERROR: missing input coordinates (-a)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
        if not self._iDb.doesTableExist( self._inCoord ):
            if not os.path.exists( self._inCoord ):
                msg = "ERROR: no file nor table for input coordinates '%s' (-a)" % ( self._inCoord )
                sys.stderr.write( "%s\n" % ( msg ) )
                self.help(); sys.exit(1)
            else:
                self._typeInCoord = "file"
        else:
            self._typeInCoord = "table"
        if self._inGenome == "":
            msg = "ERROR: missing genomic sequences (-g)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
        if not self._iDb.doesTableExist( self._inGenome ):
            if not os.path.exists( self._inGenome ):
                msg = "ERROR: no file nor table for genomic sequences '%s' (-g)" % ( self._inGenome )
                sys.stderr.write( "%s\n" % ( msg ) )
                self.help(); sys.exit(1)
            else:
                self._typeInGenome = "file"
        else:
            self._typeInGenome = "table"
        if self._inRefseqs == "":
            msg = "ERROR: missing reference sequences (-s)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
        if not self._iDb.doesTableExist( self._inRefseqs ):
            if not os.path.exists( self._inRefseqs ):
                msg = "ERROR: no file nor table for reference sequences '%s' (-s)" % ( self._inRefseqs )
                sys.stderr.write( "%s\n" % ( msg ) )
                self.help(); sys.exit(1)
            else:
                self._typeInRefseqs = "file"
        else:
            self._typeInRefseqs = "table"
        if self._formatInCoord == "":
            msg = "ERROR: missing format of input coordinates (-f)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
        if self._formatInCoord not in ["path"]:
            msg = "ERROR: unknown format '%s'" % ( self._formatInCoord )
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
        if self._data not in [ "matches", "chains" ]:
            msg = "ERROR: missing requested data (-d)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
        if self._minProp < 0 or self._minProp > 1:
            msg = "ERROR: minimum proportion outside range [0;1] (-y)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help(); sys.exit(1)
            
            
    def getListRefseqNames( self ):
        """
        Return a list with the names of reference sequences (subjects for 'path' format).
        """
        if self._refseqName != "":
            return [ self._refseqName ]
        else:
            lRefseqNames = self._tpa.getSubjectList()
            if i._verbose > 0:
                print "nb of reference sequences: %i" % ( len(lRefseqNames) )
                sys.stdout.flush()
            return lRefseqNames
        
        
    def getLengthRefseqs( self, lRefseqNames ):
        """
        Return a dictionary which keys are the names of reference sequences and values are their length.
        """
        dRefseq2Length = {}
        if self._typeInRefseqs == "table":
            for refseqName in lRefseqNames:
                dRefseq2Length[ refseqName ] = self._refseqsAdaptator.getSeqLengthFromAccession( refseqName )
        else:
            refseqFileHandler = open( self._inRefseqs, "r" )
            iBs = Bioseq()
            while True:
                iBs.reset()
                iBs.read( refseqFileHandler )
                if iBs.sequence == None:
                    break
                if iBs.header in lRefseqNames:
                    dRefseq2Length[ iBs.header ] = iBs.getLength()
                    if len(dRefseq2Length.keys()) == len(lRefseqNames):
                        break
                refseqFileHandler.close()
        return dRefseq2Length
    
    
    def saveMatches( self, refseqName, refseqLength  ):
        """
        Save all the matches of a given reference sequence into a fasta file.
        """
        outFile = "%s_matches.fa" % ( refseqName )
        outFileHandler = open( outFile, "w" )
        lPaths = self._tpa.getPathListFromSubject( refseqName )
        if self._verbose > 0:
            msg = "refseq '%s' (%i bp): %i matches" % ( refseqName, refseqLength, len(lPaths) )
            print msg; sys.stdout.flush()
        nbSaved = 0
        for iPath in lPaths:
            if self._getAnnot:
                if iPath.getLengthOnQuery() / float(refseqLength) >= self._minProp:
                    iPath.save( self._outPathFile )
        lSets = PathUtils.getSetListFromQueries( lPaths )
        for iSet in lSets:
            if self._verbose > 1:
                print iSet; sys.stdout.flush()
            if iSet.getLength() / float(refseqLength) >= self._minProp:
                bs = self._genomeAdaptator.getBioseqFromSetList( [ iSet ] )
                bs.write( outFileHandler )
                nbSaved += 1
            else:
                if self._verbose > 1:
                    print "too short"; sys.stdout.flush()
        outFileHandler.close()
        if self._verbose > 0:
            if nbSaved == 0:
                msg = "none saved"
                print msg; sys.stdout.flush()
            else:
                msg = "%i saved in file '%s'" % ( nbSaved, outFile )
                print msg; sys.stdout.flush()
        if nbSaved == 0:
            if os.path.exists( outFile ):
                os.remove( outFile )
                
                
    def saveChains( self, refseqName, refseqLength  ):
        """
        Save all the chains of matches of a given reference sequence into a fasta file.
        """
        outFile = "%s_chains.fa" % ( refseqName )
        outFileHandler = open( outFile, "a" )
        lPathnums = self._tpa.getIdListFromSubject( refseqName )
        if self._verbose > 0:
            msg = "refseq '%s': %i chains"  %( refseqName, len(lPathnums) )
            print msg; sys.stdout.flush()
        for pathnum in lPathnums:
            lPaths = self._tpa.getPathListFromId( pathnum )  # lPaths is a chain
            if self._getAnnot:
                if PathUtils.getLengthOnQueryFromPathList( lPaths ) / float(refseqLength) >= self._minProp:
                    PathUtils.writeListInFile( lPaths, self._outPathFile, "a" )
            lSets = PathUtils.getSetListFromQueries( lPaths )
            if SetUtils.getCumulLength( lSets ) / float(refseqLength) >= self._minProp:
                bs = self._genomeAdaptator.getBioseqFromSetList( lSets )
                bs.write( outFileHandler )
        outFileHandler.close()
        
        
    def start( self ):
        """
        Useful commands before running the program.
        """
        self.checkAttributes()
        if self._verbose > 0:
            print "START GetSequencesFromAnnotations.py (%s)" % ( time.strftime("%m/%d/%Y %H:%M:%S") )
            sys.stdout.flush()
        if self._typeInCoord == "file":
            tmpCoordTable = self._inCoord.replace(".","_").replace("-","_")
            self._iDb.createTable( tmpCoordTable, self._formatInCoord, self._inCoord, True, self._verbose-2 )
            self._tpa = TablePathAdaptator( self._iDb, tmpCoordTable )
        else:
            self._tpa = TablePathAdaptator( self._iDb, self._inCoord )
        if self._typeInGenome == "file":
            tmpSeqTable = self._inGenome.replace(".","_").replace("-","_")
            self._iDb.createTable( tmpSeqTable, "fasta", self._inGenome, True, self._verbose-2 )
            self._genomeAdaptator = TableSeqAdaptator( self._iDb, tmpSeqTable )
        else:
            self._genomeAdaptator = TableSeqAdaptator( self._iDb, self._inGenome )
        if self._refseqName == "file":
            pass
        else:
            self._refseqsAdaptator = TableSeqAdaptator( self._iDb, self._inRefseqs )
        if self._getAnnot:
            if self._refseqName != "":
                self._outPathFile = "%s.path" % ( self._refseqName )
            else:
                self._outPathFile = "%s.path" % ( self._inCoord )
            self._outPathHandler = open( self._outPathFile, "w" )
            
            
    def end( self ):
        """
        Useful commands before ending the program.
        """
        if self._typeInCoord == "file":
            self._iDb.dropTable( self._tpa._table, self._verbose-1 )
        if self._typeInGenome == "file":
            self._iDb.dropTable( self._genomeAdaptator._table, self._verbose-1 )
        self._iDb.close()
        if self._getAnnot:
            self._outPathHandler.close()
        if self._verbose > 0:
            print "END GetSequencesFromAnnotations.py (%s)" % ( time.strftime("%m/%d/%Y %H:%M:%S") )
            sys.stdout.flush()
            
            
    def run( self ):
        """
        Run the program.
        """
        self.start()
        
        lRefseqNames = self.getListRefseqNames()
        
        dRefseq2Length = self.getLengthRefseqs( lRefseqNames )
        
        for refseqName in lRefseqNames:
            
            if self._data == "matches":
                self.saveMatches( refseqName, dRefseq2Length[ refseqName ] )
                
            if self._data == "chains":
                self.saveChains( refseqName, dRefseq2Length[ refseqName ] )
                
        self.end()
        
        
if __name__ == "__main__":
    i = GetSequencesFromAnnotations()
    i.setAttributesFromCmdLine()
    i.run()
