#!/usr/bin/env python

##@file
# Give summary information on a TE annotation table.
# options:
#      -h: this help
#      -p: name of the table with the annotated TE copies (format='path')
#      -s: name of the table with the TE reference sequences (format='seq')
#      -g: length of the genome (in bp)
#      -o: name of the output file (default=pathTable+'_stats.txt')
#      -C: name of the configuration file
#      -v: verbosity level (default=0/1)


import sys
import getopt

from pyRepetUnit.commons.sql.DbMySql import DbMySql
from pyRepetUnit.commons.sql.TablePathAdaptator import TablePathAdaptator
from pyRepetUnit.commons.sql.TableSeqAdaptator import TableSeqAdaptator
from pyRepet.util.Stat import Stat
from repet_tools.getCumulLengthFromTEannot import getCumulLengthFromTEannot


class GiveInfoTEannot( object ):
    """
    Give summary information on a TE annotation table.
    """
    
    
    def __init__( self ):
        """
        Constructor.
        """
        self._pathTable = ""
        self._seqTable = ""
        self._genomeLength = 0
        self._outFileName = ""
        self._configFileName = ""
        self._verbose = 0
        self._db = None
        self._pathA = None
        self._seqA = None
        
        
    def help( self ):
        """
        Display the help on stdout.
        """
        print
        print "usage:",sys.argv[0].split("/")[-1],"[options]"
        print "options:"
        print "     -h: this help"
        print "     -p: name of the table with the annotated TE copies (format='path')"
        print "     -s: name of the table with the TE reference sequences (format='seq')"
        print "     -g: length of the genome (in bp)"
        print "     -o: name of the output file (default=pathTable+'_stats.txt')"
        print "     -C: name of the configuration file"
        print "     -v: verbosity level (default=0/1)"
        print
        
        
    def setAttributesFromCmdLine( self ):
        """
        Set the attributes from the command-line.
        """
        try:
            opts, args = getopt.getopt(sys.argv[1:],"hp:s:g:o:C:v:")
        except getopt.GetoptError, err:
            print str(err); self.help(); sys.exit(1)
        for o,a in opts:
            if o == "-h":
                self.help(); sys.exit(0)
            elif o == "-p":
                self._pathTable = a
            elif o == "-s":
                self._seqTable = a
            elif o == "-g":
                self._genomeLength = int(a)
            elif o == "-o":
                self._outFileName = a
            elif o == "-C":
                self._configFileName = a
            elif o == "-v":
                self._verbose = int(a)
                
                
    def checkAttributes( self ):
        """
        Check the attributes are valid before running the algorithm.
        """
        if self._pathTable == "":
            print "ERROR: missing input table"
            self.help(); sys.exit(1)
        if self._configFileName == "":
            print "ERROR: missing configuration file"
            self.help(); sys.exit(1)
        if self._outFileName == "":
            self._outFileName = "%s_stats.txt" % ( self._pathTable )
            
            
    def getCumulCoverage( self, nameTErefseq="" ):
        """
        Get the coverage of TE copies for a given family (using 'mapOp').
        @param nameTErefseq: name of a TE family ('subject_name' in the 'path' table)
        @type nameTErefseq: string
        @return: cumulative coverage
        @rtype: integer
        """
        gclft = getCumulLengthFromTEannot()
        gclft.setInputTable( self._pathTable )
        gclft.setTErefseq( nameTErefseq )
        gclft.setConfigFileName( self._configFileName )
        gclft.setClean()
        gclft._db = self._db
        gclft._tpA = self._pathA
        mapFileName = gclft.getAllSubjectsAsMapOfQueries()
        mergeFileName = gclft.mergeRanges( mapFileName )
        cumulCoverage = gclft.getCumulLength( mergeFileName )
        return cumulCoverage
    
    
    def getNbFullLengths( self, lengthTErefseq, lLengths ):
        """
        Get the number of full-lengths (95%<=L=<105%).
        @param lengthTErefseq: 
        @type lengthTErefseq: integer
        @param lLengths: list of lengths
        @type lLengthPerCopy: list of integers
        @return: cumulative coverage
        @rtype: integer
        """
        nbFullLengths = 0
        for i in lLengths:
            if i / float(lengthTErefseq ) >= 0.95 and i / float(lengthTErefseq ) <= 1.05:
                nbFullLengths += 1
        return nbFullLengths
    
    
    def getDataPerTErefseq( self, nameTErefseq, dAllTErefseqs ):
        """
        Return a dictionary with all data about a given TE reference sequence.
        """
        dOneTErefseq = { "length": 0, \
                        "cumulCoverage": 0, \
                        "nbFragments": 0, \
                        "nbFullLengthFragments": 0, \
                        "nbCopies": 0, \
                        "nbFullLengthCopies": 0, \
                        "statsIdentityPerChain": Stat(), \
                        "statsLengthPerChain": Stat(), \
                        "statsLengthPerChainPerc": Stat()
                        }
        
        lengthTErefseq = self._seqA.getSeqLengthFromAccession( nameTErefseq )
        if self._verbose > 1:
            print " length: %i bp" % ( lengthTErefseq )
        dOneTErefseq[ "length" ] = lengthTErefseq
        
        cumulCoverage = self.getCumulCoverage( nameTErefseq )  #self._pathA.getCumulPathLength_from_subject( nameTErefseq )
        if self._verbose > 1:
            print " cumulative coverage: %i bp" % ( cumulCoverage )
        dOneTErefseq[ "cumulCoverage" ] = cumulCoverage
        dAllTErefseqs[ "sumCumulCoverage" ] += cumulCoverage
        
        lLengthPerFragment = self._pathA.getPathLengthListFromSubject( nameTErefseq )
        if self._verbose > 1:
            print " nb of fragments: %i" % ( len(lLengthPerFragment) )
        dOneTErefseq[ "nbFragments" ] = len(lLengthPerFragment)
        dAllTErefseqs[ "totalNbFragments" ] += len(lLengthPerFragment)
        
        nbFullLengthFragments = self.getNbFullLengths( lengthTErefseq, lLengthPerFragment )
        if self._verbose > 1:
            print " nb full-length fragments: %i" % ( nbFullLengthFragments )
        dOneTErefseq[ "nbFullLengthFragments" ] = nbFullLengthFragments
        dAllTErefseqs[ "totalNbFullLengthFragments" ] += nbFullLengthFragments
        if nbFullLengthFragments > 0:
            dAllTErefseqs[ "nbFamWithFullLengthFragments" ] += 1
        if nbFullLengthFragments == 1:
            dAllTErefseqs[ "nbFamWithOneFullLengthFragment" ] += 1
        elif nbFullLengthFragments == 2:
            dAllTErefseqs[ "nbFamWithTwoFullLengthFragments" ] += 1
        elif nbFullLengthFragments == 3:
            dAllTErefseqs[ "nbFamWithThreeFullLengthFragments" ] += 1
        elif nbFullLengthFragments > 3:
            dAllTErefseqs[ "nbFamWithMoreThanThreeFullLengthFragments" ] += 1
            
        lLengthPerCopy = self._pathA.getChainLengthListFromSubject( nameTErefseq )
        dOneTErefseq[ "nbCopies" ] = len(lLengthPerCopy)
        dAllTErefseqs[ "totalNbCopies" ] += len(lLengthPerCopy)
        if self._verbose > 1:
            print " nb of copies: %i" % ( len(lLengthPerCopy) )
        lLengthPercPerCopy = []
        for i in lLengthPerCopy:
            dOneTErefseq[ "statsLengthPerChain" ].add( i )
            lperc = 100 * i / float(lengthTErefseq)
            lLengthPercPerCopy.append( lperc )
            dOneTErefseq[ "statsLengthPerChainPerc" ].add( lperc )
        if dOneTErefseq[ "statsLengthPerChainPerc" ].n != 0:
            dAllTErefseqs[ "statsAllCopiesMedLengthPerc" ].add( dOneTErefseq[ "statsLengthPerChainPerc" ].median() )
            if self._verbose > 1:
                print " copy length: %s" % ( dOneTErefseq[ "statsLengthPerChain" ].string() )
                
        nbFullLengthCopies = self.getNbFullLengths( lengthTErefseq, lLengthPerCopy )
        if self._verbose > 1:
            print " nb full-length copies: %i" % ( nbFullLengthCopies )
        dOneTErefseq[ "nbFullLengthCopies" ] = nbFullLengthCopies
        dAllTErefseqs[ "totalNbFullLengthCopies" ] += nbFullLengthCopies
        if nbFullLengthCopies > 0:
            dAllTErefseqs[ "nbFamWithFullLengthCopies" ] += 1
        if nbFullLengthCopies == 1:
            dAllTErefseqs[ "nbFamWithOneFullLengthCopy" ] += 1
        elif nbFullLengthCopies == 2:
            dAllTErefseqs[ "nbFamWithTwoFullLengthCopies" ] += 1
        elif nbFullLengthCopies == 3:
            dAllTErefseqs[ "nbFamWithThreeFullLengthCopies" ] += 1
        elif nbFullLengthCopies > 3:
            dAllTErefseqs[ "nbFamWithMoreThanThreeFullLengthCopies" ] += 1
            
        lIdentityPerCopy = self._pathA.getChainIdentityListFromSubject( nameTErefseq )
        for i in lIdentityPerCopy:
            dOneTErefseq[ "statsIdentityPerChain" ].add( i )
        if dOneTErefseq[ "statsIdentityPerChain" ].n != 0:
            dAllTErefseqs[ "statsAllCopiesMedIdentity" ].add( dOneTErefseq[ "statsIdentityPerChain" ].median() )
            if self._verbose > 1:
                print " copy identity: %s" % ( dOneTErefseq[ "statsIdentityPerChain" ].string() )
                
        return dOneTErefseq
    
    
    def getDataPerTErefseqAsString( self, nameTErefseq, d ):
        """
        Return a string with all data properly formatted.
        """
        string = ""
        string += "%s" % ( nameTErefseq )
        string += "\t%i" % ( d["length"] )
        string += "\t%i" % ( d["cumulCoverage"] )
        string += "\t%i" % ( d["nbFragments"] )
        string += "\t%i" % ( d["nbFullLengthFragments"] )
        string += "\t%i" % ( d["nbCopies"] )
        string += "\t%i" % ( d["nbFullLengthCopies"] )
        
        if d["statsIdentityPerChain"].n != 0:
            string += "\t%.2f" % ( d["statsIdentityPerChain"].mean() )
            string += "\t%.2f" % ( d["statsIdentityPerChain"].sd() )
            string += "\t%.2f" % ( d["statsIdentityPerChain"].min )
            string += "\t%.2f" % ( d["statsIdentityPerChain"].quantile( 0.25 ) )
            string += "\t%.2f" % ( d["statsIdentityPerChain"].median() )
            string += "\t%.2f" % ( d["statsIdentityPerChain"].quantile( 0.75 ) )
            string += "\t%.2f" % ( d["statsIdentityPerChain"].max )
        else:
            for i in range(0,7):
                string += "\tNA"
            
        if d["statsLengthPerChain"].n != 0:
            string += "\t%.2f" % ( d["statsLengthPerChain"].mean() )
            string += "\t%.2f" % ( d["statsLengthPerChain"].sd() )
            string += "\t%i" % ( d["statsLengthPerChain"].min )
            string += "\t%.2f" % ( d["statsLengthPerChain"].quantile( 0.25 ) )
            string += "\t%.2f" % ( d["statsLengthPerChain"].median() )
            string += "\t%.2f" % ( d["statsLengthPerChain"].quantile( 0.75 ) )
            string += "\t%i" % ( d["statsLengthPerChain"].max )
        else:
            for i in range(0,7):
                string += "\tNA"
                
        if d["statsLengthPerChainPerc"].n != 0:
            string += "\t%.2f" % ( d["statsLengthPerChainPerc"].mean() )
            string += "\t%.2f" % ( d["statsLengthPerChainPerc"].sd() )
            string += "\t%.2f" % ( d["statsLengthPerChainPerc"].min )
            string += "\t%.2f" % ( d["statsLengthPerChainPerc"].quantile( 0.25 ) )
            string += "\t%.2f" % ( d["statsLengthPerChainPerc"].median() )
            string += "\t%.2f" % ( d["statsLengthPerChainPerc"].quantile( 0.75 ) )
            string += "\t%.2f" % ( d["statsLengthPerChainPerc"].max )
        else:
            for i in range(0,7):
                string += "\tNA"
                
        return string
    
    
    def start( self ):
        """
        Useful commands before running the program.
        """
        self.checkAttributes()
        if self._verbose > 0:
            print "START %s" % ( type(self).__name__ ); sys.stdout.flush()
        self._db = DbMySql( cfgFileName=self._configFileName )
        if not self._db.exist( self._pathTable ):
            string = "input table '%s' doesn't exist" % ( self._pathTable )
            print "ERROR: %s" % ( string )
            sys.exit(1)
        self._pathA = TablePathAdaptator( self._db, self._pathTable )
        if not self._db.exist( self._seqTable ):
            string = "input table '%s' doesn't exist" % ( self._seqTable )
            print "ERROR: %s" % ( string )
            sys.exit(1)
        self._seqA = TableSeqAdaptator( self._db, self._seqTable )
        
        
    def end( self ):
        """
        Useful commands before ending the program.
        """
        if self._verbose > 0:
            print "END %s" % ( type(self).__name__ ); sys.stdout.flush()
            
            
    def run( self ):
        """
        Run the program.
        """
        self.start()
        outF = open( self._outFileName, "w" )
        string = "family\tlength\tcovg"
        string += "\tfrags\tfullLgthFrags\tcopies\tfullLgthCopies"  # 4 items
        string += "\tmeanId\tsdId\tminId\tq25Id\tmedId\tq75Id\tmaxId"  # 7 items
        string += "\tmeanLgth\tsdLgth\tminLgth\tq25Lgth\tmedLgth\tq75Lgth\tmaxLgth"  # 7 items
        string += "\tmeanLgthPerc\tsdLgthPerc\tminLgthPerc\tq25LgthPerc\tmedLgthPerc\tq75LgthPerc\tmaxLgthPerc"  # 7 items
        string += "\n"
        outF.write( string )
        
        lNamesTErefseq = self._seqA.getAccessionsList()
        if self._verbose > 0:
            print "nb of TE reference sequences: %i"  %( len(lNamesTErefseq) )
            sys.stdout.flush()
            
        lDistinctSubjects = self._pathA.getSubjectList()
        if self._verbose > 0:
            print "nb of distinct subjects: %i"  %( len(lDistinctSubjects) )
            sys.stdout.flush()
            
        totalCumulCoverage = self.getCumulCoverage()
        if self._verbose > 0:
            print "total cumulative coverage: %i bp" % ( totalCumulCoverage )
            if self._genomeLength > 0:
                print "TE content: %.2f%%" % ( 100 * totalCumulCoverage / float(self._genomeLength) )
                
        dAllTErefseqs = { "sumCumulCoverage": 0, \
                         "totalNbFragments": 0, \
                         "totalNbFullLengthFragments": 0, \
                         "totalNbCopies": 0, \
                         "totalNbFullLengthCopies": 0, \
                         "nbFamWithFullLengthFragments": 0, \
                         "nbFamWithOneFullLengthFragment": 0, \
                         "nbFamWithTwoFullLengthFragments": 0, \
                         "nbFamWithThreeFullLengthFragments": 0, \
                         "nbFamWithMoreThanThreeFullLengthFragments": 0, \
                         "nbFamWithFullLengthCopies": 0, \
                         "nbFamWithOneFullLengthCopy": 0, \
                         "nbFamWithTwoFullLengthCopies": 0, \
                         "nbFamWithThreeFullLengthCopies": 0, \
                         "nbFamWithMoreThanThreeFullLengthCopies": 0, \
                         "statsAllCopiesMedIdentity": Stat(), \
                         "statsAllCopiesMedLengthPerc": Stat()
                         }
        
        if self._verbose > 0:
            print "processing the %i TE families..." % ( len(lNamesTErefseq) )
            sys.stdout.flush()
        for nameTErefseq in lNamesTErefseq:
            if self._verbose > 1:
                print "processing '%s'..." % ( nameTErefseq )
                sys.stdout.flush()
                
            dOneTErefseq = self.getDataPerTErefseq( nameTErefseq, dAllTErefseqs )
            
            string = self.getDataPerTErefseqAsString( nameTErefseq, dOneTErefseq )
            
            outF.write( "%s\n" % ( string ) )
            
        outF.close()
        
        if self._verbose > 0:
            print "(sum of cumulative coverages: %i bp)" % ( dAllTErefseqs[ "sumCumulCoverage" ] )
            
            print "total nb of TE fragments: %i" % ( dAllTErefseqs[ "totalNbFragments" ] )
            
            if dAllTErefseqs[ "totalNbFragments" ] != 0:
                
                print "total nb full-length fragments: %i (%.2f%%)" % \
                ( dAllTErefseqs[ "totalNbFullLengthFragments" ], \
                  100*dAllTErefseqs[ "totalNbFullLengthFragments" ] / float(dAllTErefseqs[ "totalNbFragments" ]) )
                
                print "total nb of TE copies: %i" % ( dAllTErefseqs[ "totalNbCopies" ] )
                
                print "total nb full-length copies: %i (%.2f%%)" % \
                ( dAllTErefseqs[ "totalNbFullLengthCopies" ], \
                  100*dAllTErefseqs[ "totalNbFullLengthCopies" ] / float(dAllTErefseqs[ "totalNbCopies" ]) )
                
                print "families with full-length fragments: %i (%.2f%%)" % \
                ( dAllTErefseqs[ "nbFamWithFullLengthFragments" ], \
                  100*dAllTErefseqs[ "nbFamWithFullLengthFragments" ] / float(len(lNamesTErefseq)) )
                print " with only one full-length fragment: %i" % ( dAllTErefseqs[ "nbFamWithOneFullLengthFragment" ] )
                print " with only two full-length fragments: %i" % ( dAllTErefseqs[ "nbFamWithTwoFullLengthFragments" ] )
                print " with only three full-length fragments: %i" % ( dAllTErefseqs[ "nbFamWithThreeFullLengthFragments" ] )
                print " with more than three full-length fragments: %i" % ( dAllTErefseqs[ "nbFamWithMoreThanThreeFullLengthFragments" ] )
                
                print "families with full-length copies: %i (%.2f%%)" % \
                ( dAllTErefseqs[ "nbFamWithFullLengthCopies" ], \
                  100*dAllTErefseqs[ "nbFamWithFullLengthCopies" ] / float(len(lNamesTErefseq)) )
                print " with only one full-length copy: %i" % ( dAllTErefseqs[ "nbFamWithOneFullLengthCopy" ] )
                print " with only two full-length copies: %i" % ( dAllTErefseqs[ "nbFamWithTwoFullLengthCopies" ] )
                print " with only three full-length copies: %i" % ( dAllTErefseqs[ "nbFamWithThreeFullLengthCopies" ] )
                print " with more than three full-length copies: %i" % ( dAllTErefseqs[ "nbFamWithMoreThanThreeFullLengthCopies" ] )
                
                print "mean of median identity among all families: %.2f +- %.2f" % \
                ( dAllTErefseqs[ "statsAllCopiesMedIdentity" ].mean(), \
                  dAllTErefseqs[ "statsAllCopiesMedIdentity" ].sd() )
                
                print "mean of median length perc among all families: %.2f +- %.2f" % \
                ( dAllTErefseqs[ "statsAllCopiesMedLengthPerc" ].mean(), \
                  dAllTErefseqs[ "statsAllCopiesMedLengthPerc" ].sd() )
            
        self.end()
        
        
if __name__ == "__main__":
    i = GiveInfoTEannot()
    i.setAttributesFromCmdLine()
    i.run()
